
/**
 * @file main.cpp
 * @brief This file contains the main function for the BIAS program.
 * The BIAS program is responsible for starting and stopping the AstriDAQ and AstriMaProcessor components.
 * It reads configuration from YAML files, mysql db and command line arguments, and registers observers for data acquisition.
 * The program also handles termination signals and prints statistics before exiting.
 */
#include <Astri_DAQ.h>
#include <Astri_Redis_DAQ_Observer.h>
#include <Astri_MySQL_DAQ_Observer.h>
#include <Astri_MA_Processor.h>
#include <Astri_MA_Generic.h>

#include <Yaml_Conf.h>
#include <CL_Conf.h>
#include <Mysql_Conf.h>

#include <csignal>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <atomic>
#include <thread>
#include <mutex>

#include <ctime>
#include <iomanip>

#ifndef ASTRIDAQ_CONFIG_PATH
#define ASTRIDAQ_CONFIG_PATH std::string(std::getenv("HOME")).append("/BIAS/config/AstriDAQ")
#endif
#ifndef ASTRIPROC_CONFIG_PATH
#define ASTRIPROC_CONFIG_PATH std::string(std::getenv("HOME")).append("/BIAS/config/AstriProcs")
#endif

using namespace inaf::oasbo;
using namespace inaf::oasbo::DAQ;
using namespace inaf::oasbo::DAQ_observers;

AstriDAQ *daq;
Processors::AstriMaProcessor *procs;

void my_handler(int s) {
	time_t now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[BIAS]" << std::endl;
	daq->getMonitorPtr()->printStats();
	std::cout << "-------------------------------" << std::endl;

	now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[BIAS]\t" << "Termination..." << std::endl;

	daq->stop();
	procs->stop();
}

int main(int argc, char **argv) {

///////// DAQ CONF -------------------------------------------------------------------------------------------------
	Configurators::YamlConfigurator daqyamlconf(
			std::string(ASTRIDAQ_CONFIG_PATH).append("/config.yaml"));
	Configurators::MysqlConfigurator mysqlConf("127.0.0.1", 3306, "bias",
			"BIAS1ast01$1", "BIAS");
	Configurators::CLConfigurator clConf(argc, argv);
	daq =
			AstriDAQBuilder().configFrom(daqyamlconf)->configFrom(mysqlConf)->configFrom(
					clConf)->getDAQ();

	std::map<std::string, std::string> daqConf;
	for (auto [k, v] : daqyamlconf.getConfig())
		daqConf[k] = v;
	for (auto [k, v] : mysqlConf.getConfig())
		daqConf[k] = v;
	for (auto [k, v] : clConf.getConfig())
		daqConf[k] = v;
	time_t now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[BIAS]\t" "Starting program with DAQ params: " << std::endl;
	for (auto [k, v] : daqConf)
		std::cout << k << " : " << v << std::endl;
	std::cout << "-------------------------------" << std::endl;

	// REGISTER DAQ  OBSERVERS
	AstriRedisDAQObserver red_obs(*daq, "127.0.0.1", 6379);
	daq->registerObserver(&red_obs);
	AstriMySQLDAQObserver mysql_obs(*daq, "127.0.0.1", 3306, "bias",
			"BIAS1ast01$1", "BIAS");
	daq->registerObserver(&mysql_obs);

///////// PROCS CONF -------------------------------------------------------------------------------------------------
	Configurators::YamlConfigurator procsyamlconf(
			std::string(ASTRIPROC_CONFIG_PATH).append("/config.yaml"));
	procs =
			Processors::AstriMaProcBuilder().configFrom(procsyamlconf)->configFrom(
					mysqlConf)->configFrom(clConf)->getProcessor();
	std::map<std::string, std::string> procsConf;
	for (auto [k, v] : procsyamlconf.getConfig())
		procsConf[k] = v;
	for (auto [k, v] : mysqlConf.getConfig())
		procsConf[k] = v;
	for (auto [k, v] : clConf.getConfig())
		procsConf[k] = v;
	now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[Astri DAQ]\t" "Starting program with PROCS params: "
			<< std::endl;
	for (auto [k, v] : procsConf)
		std::cout << k << " : " << v << std::endl;
	std::cout << "-------------------------------" << std::endl;

	// CTRL-C handler
	std::signal(SIGINT, [](int signal) {
		my_handler(signal);
	});
	std::signal(SIGTERM, [](int signal) {
		my_handler(signal);
	});

	// START
	daq->start();
	procs->start();

	// WAIT FOR STOP
	while (daq->getCurrentState() != AstriDAQ::Status::STOP) {
		std::this_thread::sleep_for(std::chrono::seconds(1));
	}
	while (procs->getCurrentState()
			!= Processors::AstriMaProcessor::Status::STOP) {
		std::this_thread::sleep_for(std::chrono::seconds(1));
	}

	delete daq;
	delete procs;
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[BIAS]\t" << "Terminated normally" << std::endl;
	return EXIT_SUCCESS;
}
